//
//  xmas_tree.cpp
//
//    This sample demonstrates the use of deferred shading and HDR rendering to render
//  a Christmas tree. The tree is fractally generated, and the ornaments are all rendered
//  with point sprites using a geometry shader. The lighting is done by drawing point-sprites
//  covering the area of interest for each light, and accumulating the results into an HDR
//  buffer. Finally, that buffer is blurred and an exposure is applied to produce the final
//  image.
//
// Author: Evan Hart
// Email: sdkfeedback@nvidia.com
//
// Copyright (c) NVIDIA Corporation. All rights reserved.
////////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <map>

#include <GL/glew.h>
#include <GL/glut.h>

#define NV_REPORT_COMPILE_ERRORS
#include <nvShaderUtils.h>

#include <nvImage.h>
#include <nvSDKPath.h>

#include "tree.h"
#include "GPU_timer.h"

using std::map;

////////////////////////////////////////////////////////////////////////////////
//
// globals
//
////////////////////////////////////////////////////////////////////////////////
int width = 600, height = 800;
int dsLevel = 1;

float treeRotation = 0.0f;
nv::vec3f eyePosition(0.0f, 8.0f, 25.0f);

enum UIOption {
    OPTION_DISPLAY_TIMERS,
    OPTION_DISPLAY_INTERMEDIATES,
    OPTION_DISPLAY_VALUES,
    OPTION_USE_TREE_LIGHTS,
    OPTION_DRAW_ORNAMENTS,
    OPTION_DRAW_REFLECTIONS,
    OPTION_DRAW_FRONDS,
    OPTION_DRAW_BRANCHES,
    OPTION_ANIMATE,
    OPTION_COUNT
};
bool options[OPTION_COUNT];
map<char, UIOption> optionKeyMap;

int displayedTempBuffer = 0;

nv::SDKPath sdkPath;

Tree tree;

//
//fbo data
//
GLuint deferredShadingFBO = 0, lightAccumFBO = 0, horFilterFBO = 0, vertFilterFBO = 0;
GLuint normal = 0, pos = 0, material = 0, result = 0, hFilter = 0, vFilter = 0;
GLuint depthBuffer = 0;
GLenum hdrFormat = GL_R11F_G11F_B10F_EXT;

//
//shaders
//
GLuint basicVS, basicFS, basicProg;
GLuint defVS, defFS, defProg;
GLuint defPtVS, defPtFS, defPtGS, defPtProg;
GLuint ornamentVS, ornamentFS, ornamentGS, ornamentProg;
GLuint lightVS, lightFS, lightGS, lightProg;
GLuint reflectionFS, reflectionProg;

GLuint ppVS, displayFS, displayProg, filterFS, filterProg;

//uniform handles
GLuint filterDirection;
GLuint exposureUniform;
GLuint sigma;
GLuint blur;
GLuint level;

//values used in uniforms
float exposure = 0.125f;
float sigmaVal = 1.5f;
float blurVal = 0.4f;

//textures
GLuint dummyNormal;
GLuint frond;
GLuint frondNormal;
GLuint whiteTex;
GLuint brownTex;
GLuint cubeTex;

// timers and arrays of times used to display results
const int numTimerSamples = 120;
GPUtimer treeTimer, lightingTimer, ornamentTimer, lightsTimer, postTimer;
float treeTimes[numTimerSamples], lightingTimes[numTimerSamples], lightsTimes[numTimerSamples],
      ornamentTimes[numTimerSamples], postTimes[numTimerSamples];
int currentTime = 0;

// display lists for font rendering
GLuint fontList = 0;

////////////////////////////////////////////////////////////////////////////////
//
// Functions
//
////////////////////////////////////////////////////////////////////////////////
void createFBO();

//
//  Callback for processing ASCII key presses
//
//////////////////////////////////////////////////////////////////////
void keys( unsigned char c, int x, int y) {

    c = (unsigned char)tolower(c);

    //toggle the option, if we recognize it
    if ( optionKeyMap.find(c) != optionKeyMap.end())
        options[optionKeyMap[c]] = ! options[optionKeyMap[c]];

    switch (c) {
        case 27:
        case 'q':
            exit(-1);

        case '+':
        case '=':
            exposure *= 2.0f;
            break;
        case '-':
        case '_':
            exposure *= 0.5f;
            break;

        case 'w':
            eyePosition.z -= 0.25f;
            break;
        case 'a':
            eyePosition.x -= 0.25f;
            break;
        case 's':
            eyePosition.z += 0.25f;
            break;
        case 'd':
            eyePosition.x += 0.25f;
            break;

        case 'z':
            if (displayedTempBuffer > 0)
                displayedTempBuffer--;
            else
                displayedTempBuffer = 5;
            break;

        case 'x':
            if (displayedTempBuffer < 5)
                displayedTempBuffer++;
            else
                displayedTempBuffer = 0;
            break;

        case '[':
        case '{':
            sigmaVal -=0.25f;
            printf( "sigmaVal = %f\n", sigmaVal); 
            break;

        case ']':
        case '}':
            sigmaVal += 0.25f;
            printf( "sigmaVal = %f\n", sigmaVal);
            break;

        case '<':
        case ',':
            blurVal -= 0.05f;
            break;

        case '>':
        case '.':
            blurVal += 0.05f;
            break;

        case ';':
        case ':':
            if (dsLevel < 8)
                dsLevel++;
            createFBO();
            break;

        case '\'':
        case '"':
            if (dsLevel > 0)
                dsLevel--;
            createFBO();
            break;
    }

    glutPostRedisplay();
}

//
//  Callback for non-ASCII key presses
//
//////////////////////////////////////////////////////////////////////
void special( int k, int x, int y) {

    switch (k) {
        case GLUT_KEY_PAGE_UP:
            eyePosition.y += 0.25f;
            break;
        case GLUT_KEY_PAGE_DOWN:
            eyePosition.y -=0.25f;
            break;
    }

    glutPostRedisplay();
}

//
// Callback for the menu to change the HDR format
//
//////////////////////////////////////////////////////////////////////
void hdrMenu( int type) {
    hdrFormat = type;
    createFBO();
}

//
//  Callback for top level menu
//
//////////////////////////////////////////////////////////////////////
void menu( int val) {
    keys( val, 0, 0);
}

//
//  Draw fronds and branches with correct textures into the deferred
// shading FBO.
//
//////////////////////////////////////////////////////////////////////
void drawTreeGeometry() {
    treeTimer.startTimer();
    glUseProgram( basicProg);

    if (options[OPTION_DRAW_BRANCHES]) {
        //normal map
        glActiveTexture( GL_TEXTURE0);
        glBindTexture( GL_TEXTURE_2D, dummyNormal); // just faces straight up
        
        //diffuse map
        glActiveTexture( GL_TEXTURE1);
        glBindTexture( GL_TEXTURE_2D, brownTex);

        tree.drawBranches();
    }

    if (options[OPTION_DRAW_FRONDS]) {
        //normal map
        glActiveTexture( GL_TEXTURE0);
        glBindTexture( GL_TEXTURE_2D, frondNormal);

        //diffuse map
        glActiveTexture( GL_TEXTURE1);
        glBindTexture( GL_TEXTURE_2D, frond);

        tree.drawFronds();
    }

    treeTimer.stopTimer();
}

//
//  Draw ornament geometry into the deferred shading FBO.
//
//////////////////////////////////////////////////////////////////////
void drawOrnamentGeometry() {
    ornamentTimer.startTimer();

    if (options[OPTION_DRAW_ORNAMENTS]) {
        glUseProgram( ornamentProg);

        tree.drawOrnaments();
    }
    ornamentTimer.stopTimer();
}

//
//  Draw deferred lighting
//
//////////////////////////////////////////////////////////////////////
void drawLighting() {
    lightingTimer.startTimer();

    // bind deferred shading buffers
    glActiveTexture( GL_TEXTURE0);
    glBindTexture( GL_TEXTURE_2D, normal);
    glActiveTexture( GL_TEXTURE1);
    glBindTexture( GL_TEXTURE_2D, material);
    glActiveTexture( GL_TEXTURE2);
    glBindTexture( GL_TEXTURE_2D, pos);

    //draw a simple directional light to give base lighting
    glUseProgram( defProg);

    glBegin( GL_QUADS);
    glVertex2f( 0.0f, 0.0f);
    glVertex2f( 1.0f, 0.0f);
    glVertex2f( 1.0f, 1.0f);
    glVertex2f( 0.0f, 1.0f);
    glEnd();

    if (options[OPTION_USE_TREE_LIGHTS]) {
        //draw light points as point sprites
        glUseProgram( defPtProg);

        tree.drawLights();
    }

    if (options[OPTION_DRAW_REFLECTIONS]) {
        // add reflections to image with a full-screen pass
        glActiveTexture( GL_TEXTURE3);
        glBindTexture( GL_TEXTURE_CUBE_MAP, cubeTex);

        glUseProgram( reflectionProg);

        glBegin( GL_QUADS);
        glVertex2f( 0.0f, 0.0f);
        glVertex2f( 1.0f, 0.0f);
        glVertex2f( 1.0f, 1.0f);
        glVertex2f( 0.0f, 1.0f);
        glEnd();
        
    }


    lightingTimer.stopTimer();
}

//
//  Draw light points
//
//////////////////////////////////////////////////////////////////////
void drawLights() {
    lightsTimer.startTimer();

    if (options[OPTION_USE_TREE_LIGHTS]) {
    
        glUseProgram( lightProg);

        glEnable( GL_VERTEX_PROGRAM_POINT_SIZE);
        tree.drawLights();

        glDisable( GL_VERTEX_PROGRAM_POINT_SIZE);
    }
    lightsTimer.stopTimer();
}

//
//  Draw post-processing effects
//
//////////////////////////////////////////////////////////////////////
void drawPost() {
    postTimer.startTimer();

    
    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, horFilterFBO);

    glViewport( 0, 0, width >> dsLevel, height >> dsLevel);

    glBindTexture( GL_TEXTURE_2D, result);
    
    //downsample the result texture, using gen-mipmaps
    glGenerateMipmapEXT( GL_TEXTURE_2D);

    glUseProgram( filterProg);

    glUniform2f( filterDirection, 1.0f, 0.0f);
    glUniform1f( sigma, sigmaVal);
    glUniform1f( level, dsLevel);

    glBegin( GL_QUADS);
    glVertex2f( 0.0f, 0.0f);
    glVertex2f( 1.0f, 0.0f);
    glVertex2f( 1.0f, 1.0f);
    glVertex2f( 0.0f, 1.0f);
    glEnd();

    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, vertFilterFBO);

    glBindTexture( GL_TEXTURE_2D, hFilter);

    glUniform2f( filterDirection, 0.0f, 1.0f);
    glUniform1f( level, 0.0f);

    glBegin( GL_QUADS);
    glVertex2f( 0.0f, 0.0f);
    glVertex2f( 1.0f, 0.0f);
    glVertex2f( 1.0f, 1.0f);
    glVertex2f( 0.0f, 1.0f);
    glEnd();

    postTimer.stopTimer();
}

//
//  Draw the tree to the FBO
//
//////////////////////////////////////////////////////////////////////
void preDrawTree() {
    GLenum bufs[] = { GL_COLOR_ATTACHMENT0_EXT, GL_COLOR_ATTACHMENT1_EXT, GL_COLOR_ATTACHMENT2_EXT};

    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, deferredShadingFBO);

    glViewport( 0, 0, width, height);

    glDrawBuffers( 3, bufs);

    glClearColor( 0.0, 0.0, 0.0, 0.0);
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Draw the eye-space normals and positions for the tree geometry into the shading fbo
    drawTreeGeometry();
    
    // Draw the ornaments into the shading fbo
    drawOrnamentGeometry();

    //switch to the lighting fbo
    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, lightAccumFBO);

    // this allows occluded lights to still illuminate local geometry
    glDisable( GL_DEPTH_TEST);

    glViewport( 0, 0, width, height);

    glDrawBuffers( 1, bufs);

    glClearColor( 0.0, 0.0, 0.0, 0.0);
    glClear( GL_COLOR_BUFFER_BIT);

    glBlendFunc( GL_ONE, GL_ONE);
    glEnable( GL_BLEND);

    // draw the lighting in
    glEnable( GL_DEPTH_TEST);
    glDepthMask( GL_FALSE);

    drawLighting();

    glDepthMask( GL_TRUE);
    glDisable( GL_DEPTH_TEST);
    
    glBindTexture( GL_TEXTURE_2D, 0);
    glActiveTexture( GL_TEXTURE2);
    glBindTexture( GL_TEXTURE_2D, 0);
    glActiveTexture( GL_TEXTURE1);
    glBindTexture( GL_TEXTURE_2D, 0);
    glActiveTexture( GL_TEXTURE0);
    glBindTexture( GL_TEXTURE_2D, 0);

    //draw light points
    glEnable( GL_DEPTH_TEST);
    
    drawLights();

    glDisable( GL_BLEND);

    //downsample the framebuffer, and apply a gaussian filter
    drawPost();

    glUseProgram( 0);

    glBindTexture( GL_TEXTURE_2D, 0);

    glViewport( 0, 0, width, height);

    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0);
}

//
//  Callback for drawing the scene to the screen
//
//////////////////////////////////////////////////////////////////////
void draw() {
    glClearColor( 0.3, 0.3, 0.3, 0.0);
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glViewport( 0, 0, width, height);

    glEnable(GL_DEPTH_TEST);

    glMatrixMode( GL_PROJECTION);
    glLoadIdentity();
    gluPerspective( 40.0, (double)width/(double)height, 1.0, 100.0);
    glMatrixMode( GL_MODELVIEW);
    glLoadIdentity();

    glTranslatef( -eyePosition.x, -eyePosition.y, -eyePosition.z);
    glRotatef( treeRotation, 0.0f, 1.0f, 0.0f);

    //draw the tree to the FBO
    preDrawTree();

    //reset the viewport, because the FBO rendering will have changed it
    glViewport( 0, 0, width, height);

    //
    //draw the final tree by tonemapping it directly into the framebuffer
    //
    glBindTexture( GL_TEXTURE_2D, result);
    glActiveTexture( GL_TEXTURE1);
    glBindTexture( GL_TEXTURE_2D, vFilter);

    glUseProgram( displayProg);

    glUniform1f( exposureUniform, exposure);
    glUniform1f( blur, blurVal);

    glBegin( GL_QUADS);
    glVertex2f( 0.0f, 0.0f);
    glVertex2f( 1.0f, 0.0f);
    glVertex2f( 1.0f, 1.0f);
    glVertex2f( 0.0f, 1.0f);
    glEnd();

    glUseProgram(0);

    glBindTexture( GL_TEXTURE_2D, 0);
    glActiveTexture( GL_TEXTURE0);
    glBindTexture( GL_TEXTURE_2D, 0);

    //
    // Display overlay with intermediate results
    //
    if (options[OPTION_DISPLAY_INTERMEDIATES]) {
        const char resultTxt[] = "Result buffer";
        const char hFiltTxt[] = "Horizontal filter buffer";
        const char vFiltTxt[] = "Vertical filter buffer";
        const char normalTxt[] = "Normal buffer";
        const char positionTxt[] = "Position buffer";
        const char materialTxt[] = "Material buffer";
        const char *displayTxt;

        glMatrixMode( GL_PROJECTION);
        glLoadIdentity();
        glOrtho( 0.0, 1.0, 0.0, 1.0, 0.0, 1.0);
        glMatrixMode( GL_MODELVIEW);
        glLoadIdentity();

        glEnable(GL_TEXTURE_2D);

        switch (displayedTempBuffer) {
            case 0:
                glBindTexture( GL_TEXTURE_2D, result);
                displayTxt = resultTxt;
                break;
            case 1:
                glBindTexture( GL_TEXTURE_2D, hFilter);
                displayTxt = hFiltTxt;
                break;
            case 2:
                glBindTexture( GL_TEXTURE_2D, vFilter);
                displayTxt = vFiltTxt;
                break;
            case 3:
                glBindTexture( GL_TEXTURE_2D, normal);
                displayTxt = normalTxt;
                break;
            case 4:
                glBindTexture( GL_TEXTURE_2D, pos);
                displayTxt = positionTxt;
                break;
            case 5:
                glBindTexture( GL_TEXTURE_2D, material);
                displayTxt = materialTxt;
                break;
        };
        glColor3f( 1.0f, 1.0f, 1.0f);

        glBegin( GL_QUADS);
        glTexCoord2f( 0.0f, 0.0f);
        glVertex2f( 0.5f, 0.5f);
        glTexCoord2f( 1.0f, 0.0f);
        glVertex2f( 1.0f, 0.5f);
        glTexCoord2f( 1.0f, 1.0f);
        glVertex2f( 1.0f, 1.0f);
        glTexCoord2f( 0.0f, 1.0f);
        glVertex2f( 0.5f, 1.0f);
        glEnd();

        glDisable(GL_TEXTURE_2D);

        glBindTexture( GL_TEXTURE_2D, 0);

        // render the label
        glListBase( fontList);
        glColor3f( 1.0f, 1.0f, 1.0f);
        glWindowPos2i( width >> 1, height >> 1);
        glCallLists( GLsizei(strlen(displayTxt)), GL_BYTE, displayTxt);

    }


    //
    // Process timers and display times if appropriate
    //
    treeTimes[currentTime] = treeTimer.getLatestTime();
    ornamentTimes[currentTime] = ornamentTimer.getLatestTime();
    lightingTimes[currentTime] = lightingTimer.getLatestTime();
    lightsTimes[currentTime] = lightsTimer.getLatestTime();
    postTimes[currentTime] = postTimer.getLatestTime();

    if ( options[OPTION_DISPLAY_TIMERS]) {
        glMatrixMode( GL_PROJECTION);
        glLoadIdentity();
        glOrtho( 0.0, double(numTimerSamples), -1.0, 50.0, 0.0, 1.0);
        glMatrixMode( GL_MODELVIEW);
        glLoadIdentity();

        glBegin( GL_LINES);
        for ( int ii = numTimerSamples - 2; ii >= 0; ii--) {

            //tree rendering times in green
            glColor3f( 0.0f, 1.0f, 0.0f);
            glVertex2f( float(numTimerSamples - 2) - ii, treeTimes[ (currentTime + ii +1) % numTimerSamples]);
            glVertex2f( float(numTimerSamples - 1) - ii, treeTimes[ (currentTime + ii) % numTimerSamples]);

            //ornament rendering times in red
            glColor3f( 1.0f, 0.0f, 0.0f);
            glVertex2f( float(numTimerSamples - 2) - ii, ornamentTimes[ (currentTime + ii +1) % numTimerSamples]);
            glVertex2f( float(numTimerSamples - 1) - ii, ornamentTimes[ (currentTime + ii) % numTimerSamples]);

            //lighting rendering times in yellow
            glColor3f( 1.0f, 1.0f, 0.0f);
            glVertex2f( float(numTimerSamples - 2) - ii, lightingTimes[ (currentTime + ii +1) % numTimerSamples]);
            glVertex2f( float(numTimerSamples - 1) - ii, lightingTimes[ (currentTime + ii) % numTimerSamples]);
            
            //light rendering times in orange
            glColor3f( 1.0f, 0.7f, 0.0f);
            glVertex2f( float(numTimerSamples - 2) - ii, lightsTimes[ (currentTime + ii +1) % numTimerSamples]);
            glVertex2f( float(numTimerSamples - 1) - ii, lightsTimes[ (currentTime + ii) % numTimerSamples]);
            
            //post-processing times in blue
            glColor3f( 0.0f, 0.0f, 1.0f);
            glVertex2f( float(numTimerSamples - 2) - ii, postTimes[ (currentTime + ii +1) % numTimerSamples]);
            glVertex2f( float(numTimerSamples - 1) - ii, postTimes[ (currentTime + ii) % numTimerSamples]);
        }
        glEnd();

        //draw labels
        const char tLabel[] = "Tree cost";
        const char oLabel[] = "Ornament cost";
        const char lLabel[] = "Lighting cost";
        const char lpLabel[] = "Light point cost";
        const char ppLabel[] = "Post-processing cost";

        glListBase( fontList);

        glColor3f( 0.0f, 1.0f, 0.0f);
        glRasterPos2f( 3.0f, treeTimes[ (currentTime + 1) % numTimerSamples]+ 0.5f);
        glCallLists( GLsizei(strlen(tLabel)), GL_BYTE, tLabel);

        glColor3f( 1.0f, 0.0f, 0.0f);
        glRasterPos2f( 20.0f, ornamentTimes[ (currentTime + 1) % numTimerSamples]+ 0.5f);
        glCallLists( GLsizei(strlen(oLabel)), GL_BYTE, oLabel);

        glColor3f( 1.0f, 1.0f, 0.0f);
        glRasterPos2f( 40.0f, lightingTimes[ (currentTime + 1) % numTimerSamples]+ 0.5f);
        glCallLists( GLsizei(strlen(lLabel)), GL_BYTE, lLabel);

        glColor3f( 1.0f, 0.7f, 0.0f);
        glRasterPos2f( 60.0f, lightsTimes[ (currentTime + 1) % numTimerSamples]+ 0.5f);
        glCallLists( GLsizei(strlen(lpLabel)), GL_BYTE, lpLabel);

        glColor3f( 0.0f, 0.0f, 1.0f);
        glRasterPos2f( 80.0f, postTimes[ (currentTime + 1) % numTimerSamples]+ 0.5f);
        glCallLists( GLsizei(strlen(ppLabel)), GL_BYTE, ppLabel);
    }

    if ( options[OPTION_DISPLAY_VALUES]) {
        char buffer[256];
        int length;
        glColor3f( 1.0f, 1.0f, 1.0f);

        glListBase( fontList);

        length = sprintf( buffer, "Exposure = %1.4f", exposure);
        glWindowPos2i( 5, height - 15);
        glCallLists( length, GL_BYTE, buffer);

        length = sprintf( buffer, "Sigma (blur width) = %1.2f", sigmaVal);
        glWindowPos2i( 5, height - 30);
        glCallLists( length, GL_BYTE, buffer);

        length = sprintf( buffer, "Blur amount = %1.2f", blurVal);
        glWindowPos2i( 5, height - 45);
        glCallLists( length, GL_BYTE, buffer);

        length = sprintf( buffer, "Blur buffer size = ( %d, %d)", width >> dsLevel, height >> dsLevel);
        glWindowPos2i( 5, height - 60);
        glCallLists( length, GL_BYTE, buffer);
    }

    currentTime = (currentTime + 1) % numTimerSamples;

    glutSwapBuffers();
}

//
//
//////////////////////////////////////////////////////////////////////
void idle() {
    if (options[OPTION_ANIMATE])
        treeRotation += 0.5f;

    treeRotation = (treeRotation < 360.0f) ? treeRotation : treeRotation - 360.0f;

    glutPostRedisplay();
}

//
//
//////////////////////////////////////////////////////////////////////
void reshape( int w, int h) {

    width = w;
    height = h;

    createFBO();

    glViewport( 0, 0, w, h);
}

//
//
//////////////////////////////////////////////////////////////////////
void createFBO() {

    if ( deferredShadingFBO) {
        glDeleteFramebuffersEXT( 1, &deferredShadingFBO);
        glDeleteFramebuffersEXT( 1, &lightAccumFBO);
        glDeleteFramebuffersEXT( 1, &horFilterFBO);
        glDeleteFramebuffersEXT( 1, &vertFilterFBO);
        glDeleteTextures( 1, &normal);
        glDeleteTextures( 1, &pos);
        glDeleteTextures( 1, &material);
        glDeleteTextures( 1, &result);
        glDeleteTextures( 1, &hFilter);
        glDeleteTextures( 1, &vFilter);
        glDeleteTextures( 1, &depthBuffer);
    }

    glGenFramebuffersEXT( 1, &deferredShadingFBO);
    glGenFramebuffersEXT( 1, &lightAccumFBO);
    glGenFramebuffersEXT( 1, &horFilterFBO);
    glGenFramebuffersEXT( 1, &vertFilterFBO);
    glGenTextures( 1, &normal);
    glGenTextures( 1, &pos);
    glGenTextures( 1, &material);
    glGenTextures( 1, &result);
    glGenTextures( 1, &hFilter);
    glGenTextures( 1, &vFilter);
    glGenTextures( 1, &depthBuffer);

    //
    // Create the deferred shading FBO (normal, position, material, depth)
    //

    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, deferredShadingFBO);

    glBindTexture( GL_TEXTURE_2D, normal);
    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGB10_A2, width, height, 0, GL_RGBA, GL_FLOAT, NULL);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

    glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT1_EXT, GL_TEXTURE_2D, normal, 0);

    glBindTexture( GL_TEXTURE_2D, pos);
    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA16F_ARB, width, height, 0, GL_RGBA, GL_FLOAT, NULL);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

    glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, pos, 0);

    glBindTexture( GL_TEXTURE_2D, material);
    glTexImage2D( GL_TEXTURE_2D, 0, /*GL_RGBA8*/ GL_RGBA16F_ARB, width, height, 0, GL_RGBA, GL_FLOAT, NULL);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

    glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT2_EXT, GL_TEXTURE_2D, material, 0);

    glBindTexture( GL_TEXTURE_2D, depthBuffer);
    glTexImage2D( GL_TEXTURE_2D, 0, GL_DEPTH24_STENCIL8_EXT, width, height, 0, GL_DEPTH_COMPONENT, GL_FLOAT, NULL);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

    glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_TEXTURE_2D, depthBuffer, 0);
    glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_STENCIL_ATTACHMENT_EXT, GL_TEXTURE_2D, depthBuffer, 0);

    glBindTexture( GL_TEXTURE_2D, 0);

    if ( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) != GL_FRAMEBUFFER_COMPLETE_EXT)
        printf( "Error: bad frame buffer config\n");

    //
    // Light accum FBO
    //

    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, lightAccumFBO);

    glBindTexture( GL_TEXTURE_2D, result);
    glTexImage2D( GL_TEXTURE_2D, 0, hdrFormat, width, height, 0, GL_RGBA, GL_FLOAT, NULL);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

    glGenerateMipmapEXT( GL_TEXTURE_2D);

    glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, result, 0);

    //use the same depth buffer as the deferred shading
    glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_TEXTURE_2D, depthBuffer, 0);
    glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_STENCIL_ATTACHMENT_EXT, GL_TEXTURE_2D, depthBuffer, 0);

    glBindTexture( GL_TEXTURE_2D, 0);

    if ( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) != GL_FRAMEBUFFER_COMPLETE_EXT)
        printf( "Error: bad frame buffer config for resolve\n");

    //
    // Filter FBOs
    // 

    //horizontal filter FBO
    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, horFilterFBO);

    glBindTexture( GL_TEXTURE_2D, hFilter);
    glTexImage2D( GL_TEXTURE_2D, 0, hdrFormat, width >> dsLevel, height >> dsLevel, 0, GL_RGBA, GL_FLOAT, NULL);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

    glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, hFilter, 0);

    glBindTexture( GL_TEXTURE_2D, 0);

    if ( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) != GL_FRAMEBUFFER_COMPLETE_EXT)
        printf( "Error: bad frame buffer config for vertical filter\n");


    //vertical filter FBO
    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, vertFilterFBO);

    glBindTexture( GL_TEXTURE_2D, vFilter);
    glTexImage2D( GL_TEXTURE_2D, 0, hdrFormat, width >> dsLevel, height >> dsLevel, 0, GL_RGBA, GL_FLOAT, NULL);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

    glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, vFilter, 0);

    glBindTexture( GL_TEXTURE_2D, 0);

    if ( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) != GL_FRAMEBUFFER_COMPLETE_EXT)
        printf( "Error: bad frame buffer config for vertical filter\n");

}


//
//
//////////////////////////////////////////////////////////////////////
void loadTextures() {
    GLfloat dummy[] = { 0.5f, 0.5f, 1.0f, 0.0f};
    GLfloat white[] = { 1.0f, 1.0f, 1.0f, 1.0f};
    GLfloat brown[] = { 0.1f, 0.1f, 0.0f, 1.0f};

    glGenTextures( 1, &dummyNormal);
    glGenTextures( 1, &frondNormal);
    glGenTextures( 1, &frond);
    glGenTextures( 1, &whiteTex);
    glGenTextures( 1, &brownTex);
    glGenTextures( 1, &cubeTex);

    //
    //  simple 'out' normal (z == 1 / x,y == 0)
    //
    glBindTexture( GL_TEXTURE_2D, dummyNormal);

    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA8, 1, 1, 0, GL_RGBA, GL_FLOAT, dummy);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

    //
    // dummy white texture
    //
    glBindTexture( GL_TEXTURE_2D, whiteTex);

    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA8, 1, 1, 0, GL_RGBA, GL_FLOAT, white);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

    //
    // simple dark brown texture for tree trunka and branches
    //
    glBindTexture( GL_TEXTURE_2D, brownTex);

    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA8, 1, 1, 0, GL_RGBA, GL_FLOAT, brown);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

    //string for resolving file paths
    std::string resolved_path;

    //
    //  frond texture
    //
    nv::Image fImage;

    if ( sdkPath.getFilePath( "media/textures/FraserFirNeedles_MD_1.dds", resolved_path)) {
        if (! fImage.loadImageFromFile( resolved_path.c_str())) {
            printf( "Failed to load frond texture\n");
        }
        else {
            glBindTexture( GL_TEXTURE_2D, frond);

            glTexParameteri( GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);
            glTexImage2D( GL_TEXTURE_2D, 0, fImage.getInternalFormat(), fImage.getWidth(), fImage.getHeight(), 0, fImage.getFormat(), fImage.getType(), fImage.getLevel(0));
            glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
            glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
            glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
            glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
            glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, 16.0f);
        }
    }
    else {
        printf( "Failed to find frond texture\n");
    }

    //
    //  frond normal map
    //
    nv::Image fnImage;

    if ( sdkPath.getFilePath( "media/textures/FraserFirNeedles_MD_1_Normal.dds", resolved_path)) {
        if (! fnImage.loadImageFromFile( resolved_path.c_str())) {
            printf( "Failed to load frond normal texture\n");
        }
        else {
            glBindTexture( GL_TEXTURE_2D, frondNormal);

            glTexParameteri( GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);
            glTexImage2D( GL_TEXTURE_2D, 0, fnImage.getInternalFormat(), fnImage.getWidth(), fnImage.getHeight(), 0, fnImage.getFormat(), fnImage.getType(), fnImage.getLevel(0));
            glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
            glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
            glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
            glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
        }
    }
    else {
        printf( "Failed to find frond normal texture\n");
    }

    //
    //  stock cubemap to make balls look shiny
    //
    nv::Image cubeImage;

    if ( sdkPath.getFilePath( "media/textures/rnl_cross.dds", resolved_path)) {
        if (! cubeImage.loadImageFromFile( resolved_path.c_str())) {
            printf( "Failed to load cube texture\n");
        }
        else {
            glBindTexture( GL_TEXTURE_CUBE_MAP, cubeTex);

            // if this image does not contain pre-generated mipmaps, generate them
            if ( cubeImage.getMipLevels() == 1)
                glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_GENERATE_MIPMAP, GL_TRUE);

            //collect the parameters
            int size = cubeImage.getWidth();
            GLenum intFormat = cubeImage.getInternalFormat();
            GLenum format = cubeImage.getFormat();
            GLenum type = cubeImage.getType();

            for (int ii =0; ii < cubeImage.getMipLevels(); ii++) {
                for (int jj = 0; jj < 6; jj++) {
                    glTexImage2D( GL_TEXTURE_CUBE_MAP_POSITIVE_X + jj, ii, intFormat, size, size, 0, format, type, cubeImage.getLevel( ii, GL_TEXTURE_CUBE_MAP_POSITIVE_X  + jj));
                }
                size <<= 1;
            }
            
            glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
            glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
            glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
            glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
        }
    }
    else {
        printf( "Failed to find cube texture\n");
    }

    glBindTexture( GL_TEXTURE_2D, 0);
    glBindTexture( GL_TEXTURE_CUBE_MAP, 0);
}

//
//
//////////////////////////////////////////////////////////////////////
GLuint LinkProgram( int numShaders, ...) {
    va_list ap;

    va_start( ap, numShaders);

    GLuint program = glCreateProgram();

    for (int ii = 0; ii < numShaders; ii++) {
        GLuint shader = va_arg( ap, GLuint);

        glAttachShader( program, shader);
    }

    glLinkProgram( program);

    // check if program linked
    GLint success = 0;
    glGetProgramiv( program, GL_LINK_STATUS, &success);

    if (!success) {
        char temp[256];
        glGetProgramInfoLog( program, 256, NULL, temp);
        printf( "Failed to link program:\n%s\n", temp);
        glDeleteProgram(program);
        program = 0;
    }

    va_end(ap);

    return program;
}

//
//
//////////////////////////////////////////////////////////////////////
GLuint LinkProgramGS( GLenum inPrim, GLenum outPrim, int maxVertices, int numShaders, ...) {
    va_list ap;

    va_start( ap, numShaders);

    GLuint program = glCreateProgram();

    for (int ii = 0; ii < numShaders; ii++) {
        GLuint shader = va_arg( ap, GLuint);

        glAttachShader( program, shader);
    }

    glProgramParameteriEXT( program, GL_GEOMETRY_VERTICES_OUT_EXT, maxVertices);
    glProgramParameteriEXT( program, GL_GEOMETRY_INPUT_TYPE_EXT, inPrim);
    glProgramParameteriEXT( program, GL_GEOMETRY_OUTPUT_TYPE_EXT, outPrim);

    glLinkProgram( program);

    // check if program linked
    GLint success = 0;
    glGetProgramiv( program, GL_LINK_STATUS, &success);

    if (!success) {
        char temp[256];
        glGetProgramInfoLog( program, 256, NULL, temp);
        printf( "Failed to link program:\n%s\n", temp);
        glDeleteProgram(program);
        program = 0;
    }

    va_end(ap);

    return program;
}

//
//
//////////////////////////////////////////////////////////////////////
GLuint CompileSDKShader( GLenum shader, const char* file) {
    std::string resolved_path;

    if ( sdkPath.getFilePath( file, resolved_path)) {
        return nv::CompileGLSLShaderFromFile( shader, resolved_path.c_str());
    }
    else {
        printf( "Failed to locate shader '%s'\n", file);
    }

    return 0;
}

//
//
//////////////////////////////////////////////////////////////////////
void init() {

    //init the extensions
    glewInit();

    if ( !glewIsSupported( "GL_VERSION_2_1 GL_EXT_framebuffer_object GL_ARB_texture_float "
        "GL_EXT_gpu_shader4 GL_EXT_packed_float GL_EXT_geometry_shader4")) {
            printf( "Unable to load necessary extensions\nRequires:\n");
            printf( "  OpenGL Version 2.1\n  GL_EXT_framebuffer_object\n  GL_ARB_texture_float\n");
            printf( "  GL_EXT_packed_float\n  GL_EXT_gpu_shader4\n GL_EXT_geometry_shader4\n\n");
            exit(-1);
    }

    tree.grow();
    tree.genLights();

    createFBO();

    //
    // Compile and link all shaders
    //

    printf( "Compiling basic shaders\n");
    basicVS = CompileSDKShader( GL_VERTEX_SHADER, "src/xmas_tree/shaders/basic_vertex.glsl");
    basicFS = CompileSDKShader( GL_FRAGMENT_SHADER, "src/xmas_tree/shaders/basic_fragment.glsl");
    basicProg = LinkProgram( 2, basicVS, basicFS);

    printf( "Compiling deferred shaders\n");
    defVS = CompileSDKShader( GL_VERTEX_SHADER, "src/xmas_tree/shaders/def_shade_vertex.glsl");
    defFS = CompileSDKShader( GL_FRAGMENT_SHADER, "src/xmas_tree/shaders/def_shade_fragment.glsl");
    defProg = LinkProgram( 2, defVS, defFS);

    printf( "Compiling fullscreen shaders\n");
    ppVS = CompileSDKShader( GL_VERTEX_SHADER, "src/xmas_tree/shaders/pp_vertex.glsl");
    filterFS = CompileSDKShader( GL_FRAGMENT_SHADER, "src/xmas_tree/shaders/filter_fragment.glsl");
    filterProg = LinkProgram( 2, ppVS, filterFS);
    displayFS = CompileSDKShader( GL_FRAGMENT_SHADER, "src/xmas_tree/shaders/display_fragment.glsl");
    displayProg = LinkProgram( 2, ppVS, displayFS);
    reflectionFS = CompileSDKShader( GL_FRAGMENT_SHADER, "src/xmas_tree/shaders/reflection_fragment.glsl");
    reflectionProg = LinkProgram( 2, ppVS, reflectionFS);

    printf( "Compiling deferred point light shaders\n");
    defPtVS = CompileSDKShader( GL_VERTEX_SHADER, "src/xmas_tree/shaders/def_shade_pt_vertex.glsl");
    defPtFS = CompileSDKShader( GL_FRAGMENT_SHADER, "src/xmas_tree/shaders/def_shade_pt_fragment.glsl");
    defPtGS = CompileSDKShader( GL_GEOMETRY_SHADER_EXT, "src/xmas_tree/shaders/def_shade_pt_geometry.glsl");
    defPtProg = LinkProgramGS( GL_POINTS, GL_TRIANGLE_STRIP, 4, 3, defPtVS, defPtFS, defPtGS);
    

    printf( "Compiling ornament shaders\n");
    ornamentVS = CompileSDKShader( GL_VERTEX_SHADER, "src/xmas_tree/shaders/ornament_vertex.glsl");
    ornamentFS = CompileSDKShader( GL_FRAGMENT_SHADER, "src/xmas_tree/shaders/ornament_fragment.glsl");
    ornamentGS = CompileSDKShader( GL_GEOMETRY_SHADER_EXT, "src/xmas_tree/shaders/ornament_geometry.glsl");
    ornamentProg = LinkProgramGS( GL_POINTS, GL_TRIANGLE_STRIP, 4, 3, ornamentVS, ornamentFS, ornamentGS);

    printf( "Compiling light shaders\n");
    lightVS = CompileSDKShader( GL_VERTEX_SHADER, "src/xmas_tree/shaders/light_vertex.glsl");
    lightFS = CompileSDKShader( GL_FRAGMENT_SHADER, "src/xmas_tree/shaders/light_fragment.glsl");
    lightGS = CompileSDKShader( GL_GEOMETRY_SHADER_EXT, "src/xmas_tree/shaders/light_geometry.glsl");
    lightProg = LinkProgramGS( GL_POINTS, GL_TRIANGLE_STRIP, 4, 3, lightVS, lightFS, lightGS);
    
    //configure uniforms
    glUseProgram( basicProg);
    glUniform1i( glGetUniformLocation( basicProg, "normalTex"), 0);
    glUniform1i( glGetUniformLocation( basicProg, "baseTex"), 1);
    
    glUseProgram( defProg);
    glUniform1i( glGetUniformLocation( defProg, "materialTex"), 1);
    glUniform1i( glGetUniformLocation( defProg, "positionTex"), 2);

    glUseProgram( defPtProg);
    glUniform1i( glGetUniformLocation( defPtProg, "normalTex"), 0);
    glUniform1i( glGetUniformLocation( defPtProg, "materialTex"), 1);
    glUniform1i( glGetUniformLocation( defPtProg, "positionTex"), 2);

    glUseProgram( reflectionProg);
    glUniform1i( glGetUniformLocation( reflectionProg, "normalTex"), 0);
    glUniform1i( glGetUniformLocation( reflectionProg, "materialTex"), 1);
    glUniform1i( glGetUniformLocation( reflectionProg, "positionTex"), 2);
    glUniform1i( glGetUniformLocation( reflectionProg, "cubeTex"), 3);

    glUseProgram( displayProg);
    glUniform1i( glGetUniformLocation( displayProg, "downSampled"), 1);

    filterDirection = glGetUniformLocation( filterProg, "direction");
    exposureUniform = glGetUniformLocation( displayProg, "exposure");
    sigma = glGetUniformLocation( filterProg, "sigma");
    blur = glGetUniformLocation( displayProg, "blur");
    level = glGetUniformLocation( filterProg, "level");

    glUseProgram(0);

    //
    //  Setup the stock textures
    //
    loadTextures();


    // Init timer queroes used to track relative costs
    treeTimer.init();
    lightingTimer.init();
    lightsTimer.init();
    ornamentTimer.init();
    postTimer.init();

    //create the font display lists
    fontList = glGenLists( 128);
    for (int ii = 0; ii < 128; ii++) {
        glNewList( ii + fontList, GL_COMPILE);
        glutBitmapCharacter( GLUT_BITMAP_HELVETICA_12, ii);
        glEndList();
    }

    //default to an LEQUAL depth test
    glDepthFunc( GL_LEQUAL);
}

//
//
//////////////////////////////////////////////////////////////////////
void createMenu() {
    int topLevel = glutCreateMenu( menu);
    int hdr = glutCreateMenu( hdrMenu);

    glutAddMenuEntry( "R11F_G11F_B10F", GL_R11F_G11F_B10F_EXT);
    glutAddMenuEntry( "RGBA16F", GL_RGBA16F_ARB);
    glutAddMenuEntry( "RGBA32F", GL_RGBA32F_ARB);

    int options = glutCreateMenu( menu);

    glutAddMenuEntry( "Increase exposure [+]", '+');
    glutAddMenuEntry( "Decrease exposure [-]", '-');
    glutAddMenuEntry( "Increase sigma []]", ']');
    glutAddMenuEntry( "Decrease sigma [[]", '[');
    glutAddMenuEntry( "Increase blur amount [>]", '>');
    glutAddMenuEntry( "Decrease blur amount [<]", '<');
    glutAddMenuEntry( "Increase blur buffer size [']", '\'');
    glutAddMenuEntry( "Decrease blur buffer size [;]", ';');

    glutSetMenu( topLevel);
    glutAddSubMenu( "HDR format", hdr);
    glutAddSubMenu( "HDR options", options);
    glutAddMenuEntry( "Quit [q]", 'q');
    glutAddMenuEntry( "Stop animation [ ]", ' ');
    glutAddMenuEntry( "Toogle ornaments [o]", 'o');
    glutAddMenuEntry( "Toggle reflections [r]", 'r');
    glutAddMenuEntry( "Toggle tree lights [l]", 'l');
    glutAddMenuEntry( "Toggle fronds [f]", 'f');
    glutAddMenuEntry( "Toggle branches [b]", 'b');
    glutAddMenuEntry( "Toggle intermediates [i]", 'i');
    glutAddMenuEntry( "Toggle timers [t]", 't');
    glutAddMenuEntry( "Toggle stats [v]", 'v');
    glutAddMenuEntry( "Next temp buffer [x]", 'x');
    

    glutAttachMenu( GLUT_RIGHT_BUTTON);
}

//
//
//////////////////////////////////////////////////////////////////////
void parseTreeFile( const char *fn, Tree &tree) {
    FILE *fp = fopen( fn, "r");

    if (!fp)
        return;

    while (!feof(fp)) {
        char buf[256];

        fgets( buf, 256, fp);

        if ( buf[0] == '#') {
            //line is a comment
            continue;
        }
        else if (!strncmp( buf, "light_count", 11)) {
            tree.setLightCount( atoi( buf + 11));
        }
        else if (!strncmp( buf, "light_color", 11)) {
            float color[3];

            if ( 3 == sscanf( buf + 11, " %f %f %f", &color[0], &color[1], &color[2]) ) {
                tree.addLightColor( color[0], color[1], color[2]);
            }
        }
        else if (!strncmp( buf, "tree_height", 11)) {
            tree.setHeight( atof( buf + 11));
        }
        else if (!strncmp( buf, "tree_radius", 11)) {
            tree.setRadius(atof( buf + 11));
        }
        else if (!strncmp( buf, "ornament_material", 17)) {
            float color[4];

            if ( 4 == sscanf( buf + 17, " %f %f %f %f", &color[0], &color[1], &color[2], &color[3]) ) {
                tree.addOrnamentMaterial( color[0], color[1], color[2], color[3]);
            }
        }
        else if (!strncmp( buf, "tree_levels", 11)) {
            tree.setSegmentCount( atoi( buf + 11));
        }
        else if (!strncmp( buf, "branches_per_level", 18)) {
            tree.setBranchesPerSegment( atoi( buf + 18));
        }
        else if (!strncmp( buf, "branch_generations", 18)) {
            tree.setBranchGenerations( atoi( buf + 18));
        }
        else if (!strncmp( buf, "branch_gen_decay", 16)) {
            tree.setBranchGenerationDecayRate( atof( buf + 16));
        }
    }


    fclose(fp);
}

//
//
//////////////////////////////////////////////////////////////////////
int main( int argc, char **argv) {
    glutInit( &argc, argv);
    glutInitDisplayMode( GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);

    glutInitWindowSize( width, height);

    glutCreateWindow( "Christmas Tree Renderer");

    if (argc > 1)
        parseTreeFile( argv[1], tree);

    init();

    glutDisplayFunc( draw);
    glutIdleFunc( idle);
    glutKeyboardFunc( keys);
    glutSpecialFunc( special);
    glutReshapeFunc( reshape);

    createMenu();

    //configure the options
    optionKeyMap['t'] = OPTION_DISPLAY_TIMERS;
    options[OPTION_DISPLAY_TIMERS] = false;

    optionKeyMap['v'] = OPTION_DISPLAY_VALUES;
    options[OPTION_DISPLAY_VALUES] = false;

    optionKeyMap['l'] = OPTION_USE_TREE_LIGHTS;
    options[OPTION_USE_TREE_LIGHTS] = true;

    optionKeyMap['o'] = OPTION_DRAW_ORNAMENTS;
    options[OPTION_DRAW_ORNAMENTS] = true;

    optionKeyMap['r'] = OPTION_DRAW_REFLECTIONS;
    options[OPTION_DRAW_REFLECTIONS] = true;

    optionKeyMap['b'] = OPTION_DRAW_BRANCHES;
    options[OPTION_DRAW_BRANCHES] = true;

    optionKeyMap['f'] = OPTION_DRAW_FRONDS;
    options[OPTION_DRAW_FRONDS] = true;

    optionKeyMap['i'] = OPTION_DISPLAY_INTERMEDIATES;
    options[OPTION_DISPLAY_INTERMEDIATES] = false;
    
    optionKeyMap[' '] = OPTION_ANIMATE;
    options[OPTION_ANIMATE] = true;    


    //print the help info
    printf( "Simple_texture_buffer_object - sample showing the usage of GL_EXT_texture_buffer_object\n");
    printf( "  Commands:\n");
    printf( "    q / [ESC] - Quit the application\n");
    printf( "    [SPACE]   - Toggle continuous animation\n");
    printf( "    t         - Toggle the display of timers\n");
    printf( "    i         - Toggle display of intermediate textures\n");
    printf( "    v         - Toggle display configuration values\n");
    printf( "    o         - Toggle the drawing of ornaments\n");
    printf( "    r         - Toggle the drawing of reflections\n");
    printf( "    l         - Toggle the tree lights on/off\n");
    printf( "    f         - Toggle the tree fronds\n");
    printf( "    b         - Toggle the tree trunk and branches\n");
    printf( "    wasd      - Move camera, forward, back, left and right\n");
    printf( "    +         - Increase exposure\n");
    printf( "    -         - Decrease exposure\n");
    printf( "    z         - Cycle backward through temp buffers\n");
    printf( "    x         - Cycle forward through temp buffers\n");
    printf( "    [         - Decrease std deviation of gaussian blur (sigma)\n");
    printf( "    ]         - Decrease std deviation of gaussian blur (sigma)\n");
    printf( "    <         - Decrease blur amount\n");
    printf( "    >         - Increase blur amount\n");
    printf( "    ;         - Decrease blur buffer size\n");
    printf( "    '         - Decrease blur buffer size\n\n");




    glutMainLoop();

    return 0;
}